;;;; egg:        remote-repl
;;;; file:       remote-repl-client.scm
;;;; author:     elf <elf@ephemeral.net>
;;;; date:       16 Aug 2008
;;;; purpose:    client component of the remote repl
;;;;
;;;; licence:    BSD (see LICENCE)
;;;; copyright:  Copyright (C) 2008, Elf
;;;;             All rights reserved.
;;;;
;;;; changelog:  20080816 [elf]  Initial release.
;;;;             20080817 [elf]  Fix for read timeouts. (reported by certainty)
;;;;             20080817 [elf]  Bumped version to 1.1.0
;;;;             20080817 [elf]  Compilation warnings cleaned up.
;;;;             20080817 [elf]  Bumped version to 1.1.1
;;;;             20080817 [elf]  Bumped version to 1.1.2 
;;;;             20080817 [elf]  Removed timeout due to conflict with server.
;;;;                             Fixed output behaviour.
;;;;                             Removed (disable-interrupts) declaration.
;;;;             20080817 [elf]  Bumped version to 1.1.3
;;;;



;;; chicken library loading


(require-extension tcp)       ; tcp sockets
(require-extension srfi-18)   ; lightweight thread library



;;; chicken compile-time directives


(eval-when (compile)
    (declare
        (uses library extras eval tcp srfi-18)
        (bound-to-procedure
            rrepl-client-new
            rrepl-client?
            rrepl-client:read
            rrepl-client:read!
            rrepl-client:print
            rrepl-client:print!
            rrepl-client:close
            rrepl-client:close!
            rrepl-client:inport
            rrepl-client:inport!
            rrepl-client:outport
            rrepl-client:outport!
            rrepl-client:active
            rrepl-client:active!
            print-default
            auth-default
            close
            condition-msg
            rrepl-client-connect
            rrepl-client-send
            rrepl-client-close
        )
        (export
            rrepl-client?
            rrepl-client:read
            rrepl-client:print
            rrepl-client:close
            rrepl-client:inport
            rrepl-client:outport
            rrepl-client:active
            rrepl-client-connect
            rrepl-client-send
            rrepl-client-close
            )
        (emit-exports "remote-repl-client.exports")
        (unused
            rrepl-client:read!
            rrepl-client:print!
            rrepl-client:close!
            )
        (inline)
        (inline-limit 100)
        (lambda-lift)
        (no-bound-checks)
        (no-procedure-checks)
        (standard-bindings)
        (extended-bindings)
        (usual-integrations)
    ))



;;; data structures


;; record: rrepl-client
;; structure containing client (remote controller) data
;; fields:
;;    reader   (read)    :  proc to read a line from input port (iport)
;;    printer  (print)   :  proc to print (send) to output port (expr oport)
;;    close    (close)   :  proc to close client ports (client)
;;    inport   (inport)  :  input socket port
;;    outport  (outport) :  output socket port
;;    active   (active)  :  boolean, is client active?
(define-record-type rrepl-client
    (rrepl-client-new reader printer close inport outport active)
    rrepl-client?
        (reader     rrepl-client:read       rrepl-client:read!)
        (printer    rrepl-client:print      rrepl-client:print!)
        (close      rrepl-client:close      rrepl-client:close!)
        (inport     rrepl-client:inport     rrepl-client:inport!)
        (outport    rrepl-client:outport    rrepl-client:outport!)
        (active     rrepl-client:active     rrepl-client:active!)
    )



;; generic and default procedures


;; (print-default EXPRESSION OUTPUT-PORT)
;; default handler for sending data (newline is important!)
(define (print-default expr port)
    (write expr port)
    (newline port))

;; (auth-default CLIENT-RECORD)
;; default handler for authentication (do nothing and return client record)
(define (auth-default cli)
    cli)

;; (close INPUT-PORT-CLOSE-PROC OUTPUT-PORT-CLOSE-PROC)
;; wrapper for cleanup when closing client ports
(define (close close-in close-out)
    (lambda (cli)
        (condition-case
            (close-in (rrepl-client:inport cli))
                (var ()    #f))
        (condition-case
            (close-out (rrepl-client:outport cli))
                (var ()    #f))
        (rrepl-client:inport! cli #f)
        (rrepl-client:outport! cli #f)
        (rrepl-client:active! cli #f)))

;; (condition-msg CONDITION)
;; message component of exception
(define condition-msg    (condition-property-accessor 'exn 'message))



;;; high-level interface


;; rrepl-client-connect
;; connect to a remote-repl server
;; args:
;;    host       :  string hostname/ip where server resides
;;    port       :  integer port number where server resides
;;    connect    :  proc to perform socket connection (host port)
;;    auth       :  proc to handle authentication (client)
;;    reader     :  proc to read a line from input port (iport)
;;    printer    :  proc to send data to output port (expr oport)
;;    close-in   :  proc to close input port (iport)
;;    close-out  :  proc to close output port (oport)
(define (rrepl-client-connect host port #!key (connect     tcp-connect)
                                              (auth        auth-default)
                                              (reader      read-line)
                                              (printer     print-default)
                                              (close-in    close-input-port)
                                              (close-out   close-output-port))
    (##sys#check-string host 'rrepl-client-connect)
    (##sys#check-range port 1 32767 'rrepl-client-connect)
    (call-with-values
        (lambda ()
            (connect host port))
        (lambda (i o)
            (auth (rrepl-client-new reader
                                    printer
                                    (close close-in close-out)
                                    i
                                    o
                                    #t)))))

;; (rrepl-client-send CLIENT-RECORD EXPRESSION)
;; send data for evaluation on remote server and get result
(define (rrepl-client-send cli expr)
    (##sys#check-structure cli 'rrepl-client 'rrepl-client-send)
    (if (rrepl-client:active cli)
        (begin
            (condition-case
                ((rrepl-client:print cli) expr (rrepl-client:outport cli))
                    (var ()
                        ((rrepl-client:close cli) cli)
                        (display "write error - closing\n")))
            (if (rrepl-client:active cli)
                (let ploop ((t   #t)
                            (r   ##sys#undefined-value))
                    (condition-case
                        (set! r
                            ((rrepl-client:read cli) (rrepl-client:inport cli)))
                            (var ()
                                (if (string=? (condition-msg var)
                                              "read operation timed out")
                                    (set! t #f)
                                    (begin
                                        ((rrepl-client:close cli) cli)
                                        (display "read error - closing\n")))))
                    (if t
                        r
                        (ploop #t ##sys#undefined-value)))))
        (display "client is not active\n")))

;; (rrepl-client-close CLIENT-RECORD)
;; shut down a client connection
(define (rrepl-client-close cli)
    (##sys#check-structure cli 'rrepl-client 'rrepl-client-close)
    (if (rrepl-client:active cli)
        (begin
            (display (rrepl-client-send cli '(rrepl-session-close)))
            (newline)
            ((rrepl-client:close cli) cli))
        #t))

