;;;; egg:        remote-repl
;;;; file:       remote-repl-server.scm
;;;; author:     elf <elf@ephemeral.net>
;;;; date:       16 Aug 2008
;;;; purpose:    server component of the remote repl
;;;;
;;;; licence:    BSD (see LICENCE)
;;;; copyright:  Copyright (C) 2008, Elf
;;;;             All rights reserved.
;;;;
;;;; changelog:  20080816 [elf]  Initial release.
;;;;             20080817 [elf]  Fix for session print timeout.
;;;;             20080817 [elf]  Fix for sleep(2) blocking behaviour and 
;;;;                             main thread blocking.  (reported by certainty)
;;;;             20080817 [elf]  Bumped version to 1.1.0
;;;;             20080817 [elf]  Fix for sleep(2) behaviour in compiled code.
;;;;             20080817 [elf]  sleep (from unit posix) redefined to 
;;;;                             thread-sleep! (from unit srfi-18).
;;;;             20080817 [elf]  Compilation warnings cleaned up.
;;;;             20080817 [elf]  Bumped version to 1.1.1
;;;;             20080817 [elf]  Adjusted thread-quantum for primordial.
;;;;             20080817 [elf]  Bumped version to 1.1.2
;;;;             20080817 [elf]  Adjusted thread-quanta.
;;;;                             Removed (disable-interrupts) declaration.
;;;;             20080817 [elf]  Bumped version to 1.1.3
;;;;



;;; chicken library loading


(require-extension posix)     ; POSIX wrapper library
(require-extension tcp)       ; tcp sockets
(require-extension srfi-1)    ; list library
(require-extension srfi-18)   ; lightweight thread library



;;; chicken compile-time directives


(eval-when (compile)
    (declare
        (uses library extras eval tcp posix srfi-1 srfi-18)
        (bound-to-procedure
            rrepl-server-new
            rrepl-server?
            rrepl-server:close-listen
            rrepl-server:close-listen!
            rrepl-server:read
            rrepl-server:read!
            rrepl-server:eval
            rrepl-server:eval!
            rrepl-server:print
            rrepl-server:print!
            rrepl-server:close
            rrepl-server:close!
            rrepl-server:listener
            rrepl-server:listener!
            rrepl-server:listen-thr
            rrepl-server:listen-thr!
            rrepl-server:remotes
            rrepl-server:remotes!
            rrepl-server:accepting
            rrepl-server:accepting!
            rrepl-server:active
            rrepl-server:active!
            rrepl-session-new
            rrepl-session?
            rrepl-session:inport
            rrepl-session:inport!
            rrepl-session:outport
            rrepl-session:outport!
            rrepl-session:thread
            rrepl-session:thread!
            rrepl-session:extra
            rrepl-session:extra!
            rrepl-session:active
            rrepl-session:active!
            close-listen
            auth-default
            eval-default
            close
            condition-msg
            condition-loc
            condition-arg
            serialise-exn
            make-session-thread
            make-listener-thread
            rrepl-server-start
            rrepl-server-close
            )
        (export
            rrepl-server?
            rrepl-server:close-listen
            rrepl-server:read
            rrepl-server:eval
            rrepl-server:print
            rrepl-server:close
            rrepl-server:listener
            rrepl-server:listen-thr
            rrepl-server:remotes
            rrepl-server:accepting
            rrepl-server:accepting!
            rrepl-server:active
            rrepl-server:active!
            rrepl-session?
            rrepl-session:inport
            rrepl-session:outport
            rrepl-session:thread
            rrepl-session:extra
            rrepl-session:extra!
            rrepl-session:active
            rrepl-session:active!
            rrepl-server-start
            rrepl-server-close
            sleep
            )
        (emit-exports "remote-repl-server.exports")
        (unused
            rrepl-server:close-listen!
            rrepl-server:read!
            rrepl-server:eval!
            rrepl-server:print!
            rrepl-server:close!
            )
        (inline)
        (inline-limit 100)
        (lambda-lift)
        (no-bound-checks)
        (no-procedure-checks)
        (standard-bindings)
        (extended-bindings)
        (usual-integrations)
    ))



;;; system overrides


;; sleep -> thread-sleep! to prevent indefinite blocking
(define sleep    thread-sleep!)



;;; data structures


;; record: rrepl-server
;; structure containing general server data
;; fields:
;;    clisten    (close-listen) : proc to close the listener port (serv)
;;    reader     (read)         : proc to read from the input port (iport)
;;    evaler     (eval)         : proc to eval locally (expr sess)
;;    printer    (print)        : proc to print to the output port (expr oport)
;;    close      (close)        : proc to close the session ports (serv sess)
;;    listener   (listener)     : listener port
;;    lthread    (listen-thr)   : listener thread
;;    rthreads   (remotes)      : list of active sessions
;;    accepting  (accepting)    : boolean, is listener thread active?
;;    active     (active)       : boolean, is server running?
(define-record-type rrepl-server
    (rrepl-server-new clisten reader evaler printer close
                      listener lthread rthreads accepting active)
    rrepl-server?
        (clisten      rrepl-server:close-listen    rrepl-server:close-listen!)
        (reader       rrepl-server:read            rrepl-server:read!)
        (evaler       rrepl-server:eval            rrepl-server:eval!)
        (printer      rrepl-server:print           rrepl-server:print!)
        (close        rrepl-server:close           rrepl-server:close!)
        (listener     rrepl-server:listener        rrepl-server:listener!)
        (lthread      rrepl-server:listen-thr      rrepl-server:listen-thr!)
        (rthreads     rrepl-server:remotes         rrepl-server:remotes!)
        (accepting    rrepl-server:accepting       rrepl-server:accepting!)
        (active       rrepl-server:active          rrepl-server:active!)
    )

;; record: rrepl-session
;; structure containing per-session data
;; fields:
;;    inport  :  input socket port
;;    outport :  output socket port
;;    thread  :  thread containing session repl
;;    extra   :  user-definable data
;;    active  :  boolean, is session active?
(define-record-type rrepl-session
    (rrepl-session-new inport outport thread extra active)
    rrepl-session?
        (inport     rrepl-session:inport     rrepl-session:inport!)
        (outport    rrepl-session:outport    rrepl-session:outport!)
        (thread     rrepl-session:thread     rrepl-session:thread!)
        (extra      rrepl-session:extra      rrepl-session:extra!)
        (active     rrepl-session:active     rrepl-session:active!)
    )



;;; generic and default procedures


;; (close-listen SOCKET-LISTENER-CLOSE-PROC)
;; wrapper for cleanup when closing listener port
(define (close-listen clisten)
    (lambda (serv)
        (condition-case
            (clisten (rrepl-server:listener serv))
                (var ()    #f))
        (rrepl-server:listener! serv #f)
        (rrepl-server:accepting! serv #f)))

;; (auth-default SERVER-RECORD SESSION-RECORD)
;; default handler for authentication (do nothing and return session)
(define (auth-default serv sess)
    sess)

;; (eval-default EXPRESSION SESSION-RECORD)
;; default handler for evaluating expressions (call eval)
(define (eval-default expr sess)
    (eval expr))

;; (print-default EXPRESSION OUTPUT-PORT)
;; default handler/wrapper for output (newline is important!)
(define (print-default expr port)
    (write expr port)
    (newline port))

;; (close INPUT-PORT-CLOSE-PROC OUTPUT-PORT-CLOSE-PROC)
;; wrapper for cleanup when closing a session
(define (close close-in close-out)
    (lambda (serv sess)
        (condition-case
            (close-in (rrepl-session:inport sess))
                (var ()    #f))
        (condition-case
            (close-out (rrepl-session:outport sess))
                (var ()    #f))
        (rrepl-session:inport! sess #f)
        (rrepl-session:outport! sess #f)
        (rrepl-session:active! sess #f)
        (rrepl-server:remotes! serv (delete sess (rrepl-server:remotes serv)))))



;;; repl error printers


;; (condition-msg CONDITION)
;; message component of exception
(define condition-msg    (condition-property-accessor 'exn 'message))

;; (condition-loc CONDITION)
;; location (caller) component of exception
(define condition-loc    (condition-property-accessor 'exn 'location))

;; (condition-arg CONDITION)
;; arguments component of exception
(define condition-arg    (condition-property-accessor 'exn 'arguments))

;; (serialise-exn CONDITION)
;; create output string from an error in eval stage
(define (serialise-exn e)
    (conc "condition: " (##sys#slot e 1)
          " - " (condition-msg e)
          (if (condition-arg e)
              (conc "   args: " (condition-arg e))
              "")
          (if (condition-loc e)
              (conc "   in procedure " (condition-loc e))
              "")))



;;; thread procedures


;; (make-session-thread SERVER-RECORD SESSION-RECORD)
;; create a new session thread
(define (make-session-thread serv sess)
    (make-thread
        (lambda ()
            (define (rrepl-session-close)
                (rrepl-session:active! sess 1)
                "shutting down session...")
            (let loop ()
                (if (and (rrepl-server:active serv)
                         (rrepl-session:active sess))
                    (let* ((t   #t)
                           (r   (condition-case
                                    ((rrepl-server:read serv)
                                      (rrepl-session:inport sess))
                                        (var ()
                                            (if (string=?
                                                    (condition-msg var)
                                                    "read operation timed out")
                                                (set! t #f)
                                                ((rrepl-server:close serv)
                                                    serv sess)))))
                           (e   (and t
                                     (rrepl-session:active sess)
                                     (rrepl-server:active serv)
                                     (if (equal? r '(rrepl-session-close))
                                         (rrepl-session-close)
                                         (condition-case
                                             ((rrepl-server:eval serv) r sess)
                                                 (var ()
                                                     (serialise-exn var))))))
                           (p   (and t
                                     (rrepl-session:active sess)
                                     (let ploop ((s   #t))
                                         (condition-case
                                             ((rrepl-server:print serv)
                                               e
                                               (rrepl-session:outport sess))
                                                 (var ()
                                                     (if (string=?
                                                             (condition-msg var)
                                                             "write operation timed out")
                                                         (set! s #f)
                                                         ((rrepl-server:close serv)
                                                             serv sess))))
                                         (or s
                                             (ploop #t))))))
                        (and (eq? (rrepl-session:active sess) 1)
                             (rrepl-session:active! sess #f))
                        (thread-yield!)
                        (loop))
                    ((rrepl-server:close serv) serv sess))))))

;; (make-listener-thread SERVER-RECORD SOCKET-ACCEPT-PROC AUTH-PROC)
;; create a new listener thread
(define (make-listener-thread serv accept auth)
    (make-thread
        (lambda ()
            (let loop ()
                (if (and (rrepl-server:accepting serv)
                         (rrepl-server:active serv))
                    (let* ((s   (call-with-values
                                    (lambda ()
                                        (condition-case
                                            (accept
                                              (rrepl-server:listener serv))
                                                (val ()
                                                    (values #f #f))))
                                    (lambda (i o)
                                        (and (rrepl-server:active serv)
                                             (rrepl-server:accepting serv)
                                             i
                                             (auth serv
                                                  (rrepl-session-new
                                                      i
                                                      o
                                                      #f
                                                      #f
                                                      #t))))))
                           (t   (and s
                                     (rrepl-server:active serv)
                                     (rrepl-server:accepting serv)
                                     (make-session-thread serv s))))
                        (and t
                             (rrepl-server:active serv)
                             (rrepl-server:accepting serv)
                             (rrepl-session:thread! s t)
                             (rrepl-server:remotes! serv
                                 (cons s (rrepl-server:remotes serv)))
                             (thread-quantum-set! t
                                 (+ (tcp-read-timeout)
                                    (quotient (tcp-read-timeout) 5)))
                             (thread-start! t))
                        (thread-yield!)
                        (loop))
                    ((rrepl-server:close-listen serv) serv))))))



;;; high-level interface


;; rrepl-server-start
;; starts a new remote repl server
;; args:
;;    port           : integer, port to listen for connections
;;    listen         : proc to listen on a port (port backlog host)
;;    accept         : proc to accept connections (listener-port)
;;    clisten        : proc to close listener port (listener-port)
;;    auth           : proc to authenticate connections (serv sess)
;;    host           : string hostname/ip to only allow connections from
;;    backlog        : integer, number of pending connections for listener
;;    timeout-read   : integer time in ms for read timeout
;;    timeout-write  : integer time in ms for write timeout
;;    timeout-accept : integer time in ms for accept timeout
;;    reader         : proc to read from input port (iport)
;;    evaler         : proc to eval locally (expr sess)
;;    printer        : proc to print to output port (expr oport)
;;    close-in       : proc to close input port (iport)
;;    close-out      : proc to close output port (oport)
(define (rrepl-server-start port #!key (listen           tcp-listen)
                                       (accept           tcp-accept)
                                       (clisten          tcp-close)
                                       (auth             auth-default)
                                       (host             #f)
                                       (backlog          4)
                                       (timeout-read     50000)
                                       (timeout-write    50000)
                                       (timeout-accept   50000)
                                       (reader           read)
                                       (evaler           eval-default)
                                       (printer          print-default)
                                       (close-in         close-input-port)
                                       (close-out        close-output-port))
    (##sys#check-range port 1 32767 'rrepl-server-start)
    (tcp-accept-timeout timeout-accept)
    (tcp-read-timeout timeout-read)
    (tcp-write-timeout timeout-write)
    (thread-quantum-set! (current-thread)
        (max (thread-quantum (current-thread))
             (fx+ (fx* timeout-read 2) (fx/ timeout-read 2))))
    (let* ((s   (rrepl-server-new (close-listen clisten)
                                  reader
                                  evaler
                                  printer
                                  (close close-in close-out)
                                  (listen port backlog host)
                                  #f
                                  '()
                                  #t
                                  #t))
           (t   (make-listener-thread s accept auth)))
        (rrepl-server:listen-thr! s t)
        (thread-quantum-set! t (fx+ timeout-accept 4000))
        (thread-start! t)
        s))

;; (rrepl-server-close SERVER-RECORD)
;; shut down the server and all connections
(define (rrepl-server-close serv)
    (##sys#check-structure serv 'rrepl-server 'rrepl-server-close)
    (rrepl-server:active! serv #f)
    (thread-join! (rrepl-server:listen-thr serv))
    (for-each
        (lambda (x)
            (thread-join! (rrepl-session:thread x)))
        (rrepl-server:remotes serv))
    #t)

